import dagster_pandas as dagster_pd
from dagster_dask import dask_executor

from dagster import (
    ExecutionTargetHandle,
    InputDefinition,
    ModeDefinition,
    execute_pipeline,
    file_relative_path,
    pipeline,
    seven,
    solid,
)
from dagster.core.definitions.executor import default_executors
from dagster.core.instance import DagsterInstance
from dagster.core.test_utils import nesting_composite_pipeline


@solid
def simple(_):
    return 1


@pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
def dask_engine_pipeline():
    return simple()


def test_execute_on_dask():
    with seven.TemporaryDirectory() as tempdir:
        result = execute_pipeline(
            ExecutionTargetHandle.for_pipeline_python_file(
                __file__, 'dask_engine_pipeline'
            ).build_pipeline_definition(),
            environment_dict={
                'storage': {'filesystem': {'config': {'base_dir': tempdir}}},
                'execution': {'dask': {'config': {'timeout': 30}}},
            },
            instance=DagsterInstance.local_temp(),
        )
        assert result.result_for_solid('simple').output_value() == 1


def dask_composite_pipeline():
    return nesting_composite_pipeline(
        6, 2, mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])]
    )


def test_composite_execute():
    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_python_file(
            __file__, 'dask_composite_pipeline'
        ).build_pipeline_definition(),
        environment_dict={
            'storage': {'filesystem': {}},
            'execution': {'dask': {'config': {'timeout': 30}}},
        },
        instance=DagsterInstance.local_temp(),
    )
    assert result.success


@solid(input_defs=[InputDefinition('df', dagster_pd.DataFrame)])
def pandas_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
def pandas_pipeline():
    return pandas_solid()


def test_pandas_dask():
    environment_dict = {
        'solids': {
            'pandas_solid': {
                'inputs': {'df': {'csv': {'path': file_relative_path(__file__, 'ex.csv')}}}
            }
        }
    }

    result = execute_pipeline(
        ExecutionTargetHandle.for_pipeline_python_file(
            __file__, pandas_pipeline.name
        ).build_pipeline_definition(),
        environment_dict={
            'storage': {'filesystem': {}},
            'execution': {'dask': {'config': {'timeout': 30}}},
            **environment_dict,
        },
        instance=DagsterInstance.local_temp(),
    )

    assert result.success



import dask
import dask.distributed
from dagster_graphql.client.mutations import execute_execute_plan_mutation

from dagster import check, seven
from dagster.core.engine.engine_base import Engine
from dagster.core.events import DagsterEvent
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.plan import ExecutionPlan
from dagster.utils import frozentags
from dagster.utils.net import is_local_uri

from .config import DaskConfig

# Dask resource requirements are specified under this key
DASK_RESOURCE_REQUIREMENTS_KEY = 'dagster-dask/resource_requirements'


def query_on_dask_worker(
    handle, variables, dependencies, instance_ref=None
):  # pylint: disable=unused-argument
    '''Note that we need to pass "dependencies" to ensure Dask sequences futures during task
    scheduling, even though we do not use this argument within the function.
    '''
    return execute_execute_plan_mutation(handle, variables, instance_ref=instance_ref)


def get_dask_resource_requirements(tags):
    check.inst_param(tags, 'tags', frozentags)
    req_str = tags.get(DASK_RESOURCE_REQUIREMENTS_KEY)
    if req_str is not None:
        return seven.json.loads(req_str)

    return {}


class DaskEngine(Engine):  # pylint: disable=no-init
    @staticmethod
    def execute(pipeline_context, execution_plan):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)

        dask_config = pipeline_context.executor_config

        check.param_invariant(
            isinstance(pipeline_context.executor_config, DaskConfig),
            'pipeline_context',
            'Expected executor_config to be DaskConfig got {}'.format(
                pipeline_context.executor_config
            ),
        )

        # Checks to ensure storage is compatible with Dask configuration
        storage = pipeline_context.environment_dict.get('storage')
        check.invariant(storage.keys(), 'Must specify storage to use Dask execution')

        check.invariant(
            pipeline_context.instance.is_persistent,
            'Dask execution requires a persistent DagsterInstance',
        )

        # Remote
        if dask_config.address and not is_local_uri(dask_config.address):
            check.invariant(
                storage.get('s3') or storage.get('gcs'),
                'Must use S3 or GCS storage with non-local Dask address {dask_address}'.format(
                    dask_address=dask_config.address
                ),
            )
        # Local
        else:
            check.invariant(
                not storage.get('in_memory'),
                'Cannot use in-memory storage with Dask, use filesystem, S3, or GCS',
            )

        step_levels = execution_plan.execution_step_levels()

        pipeline_name = pipeline_context.pipeline_def.name

        instance = pipeline_context.instance

        with dask.distributed.Client(**dask_config.build_dict(pipeline_name)) as client:
            execution_futures = []
            execution_futures_dict = {}

            for step_level in step_levels:
                for step in step_level:
                    # We ensure correctness in sequencing by letting Dask schedule futures and
                    # awaiting dependencies within each step.
                    dependencies = []
                    for step_input in step.step_inputs:
                        for key in step_input.dependency_keys:
                            dependencies.append(execution_futures_dict[key])

                    environment_dict = dict(
                        pipeline_context.environment_dict, execution={'in_process': {}}
                    )
                    variables = {
                        'executionParams': {
                            'selector': {'name': pipeline_name},
                            'environmentConfigData': environment_dict,
                            'mode': pipeline_context.mode_def.name,
                            'executionMetadata': {'runId': pipeline_context.pipeline_run.run_id},
                            'stepKeys': [step.key],
                        }
                    }

                    dask_task_name = '%s.%s' % (pipeline_name, step.key)

                    future = client.submit(
                        query_on_dask_worker,
                        pipeline_context.execution_target_handle,
                        variables,
                        dependencies,
                        instance.get_ref(),
                        key=dask_task_name,
                        resources=get_dask_resource_requirements(step.tags),
                    )

                    execution_futures.append(future)
                    execution_futures_dict[step.key] = future

            # This tells Dask to awaits the step executions and retrieve their results to the
            # master
            for future in dask.distributed.as_completed(execution_futures):
                for step_event in future.result():
                    check.inst(step_event, DagsterEvent)

                    yield step_event



from dagster import Bool, Field, Int, String
from dagster.core.definitions.executor import check_cross_process_constraints, executor

from .config import DaskConfig


@executor(
    name='dask',
    config={
        'address': Field(
            String,
            is_required=False,
            description='The address of a `Scheduler` server, e.g., `\'127.0.0.1:8786\'`.',
        ),
        'timeout': Field(
            Int,
            is_required=False,
            description='Timeout duration for initial connection to the scheduler.',
        ),
        'scheduler_file': Field(
            String,
            is_required=False,
            description='Path to a file with scheduler information if available.',
        ),
        'direct_to_workers': Field(
            Bool,
            is_required=False,
            description='Whether or not to connect directly to the workers, or to ask the '
            'scheduler to serve as intermediary.',
        ),
        'heartbeat_interval': Field(
            Int,
            is_required=False,
            description='Time in milliseconds between heartbeats to scheduler.',
        ),
    },
)
def dask_executor(init_context):
    '''Dask-based executor.

    If the Dask executor is used without providing executor-specific config, a local Dask cluster
    will be created (as when calling :py:class:`dask.distributed.Client() <dask:distributed.Client>`
    without specifying the scheduler address).

    The Dask executor optionally takes the following config:

    .. code-block::

        {
            address?: '127.0.0.1:8786',  # The address of a Dask scheduler
            timeout?: 5,  # Timeout duration for initial connection to the scheduler
            scheduler_file?: '/path/to/file'  # Path to a file with scheduler information
            # Whether to connect directly to the workers, or ask the scheduler to serve as
            # intermediary
            direct_to_workers?: False,
            heartbeat_interval?: 1000,  # Time in milliseconds between heartbeats to scheduler
        }

    If you'd like to configure a dask executor in addition to the
    :py:class:`~dagster.default_executors`, you should add it to the ``executor_defs`` defined on a
    :py:class:`~dagster.ModeDefinition` as follows:

    .. code-block:: python

        from dagster import ModeDefinition, default_executors, pipeline
        from dagster_dask import dask_executor

        @pipeline(mode_defs=[ModeDefinition(executor_defs=default_executors + [dask_executor])])
        def dask_enabled_pipeline():
            pass

    '''
    check_cross_process_constraints(init_context)

    return DaskConfig(**init_context.executor_config)



from collections import namedtuple

from dagster import check
from dagster.core.execution.config import ExecutorConfig


class DaskConfig(
    namedtuple('DaskConfig', 'address timeout scheduler_file direct_to_workers heartbeat_interval'),
    ExecutorConfig,
):
    '''DaskConfig - configuration for the Dask execution engine

    Params:
        address (Optional[str]): The address of a `Scheduler` server, e.g., `'127.0.0.1:8786'`.
        timeout (Optional[int]): Timeout duration for initial connection to the scheduler.
        scheduler_file (Optional[str]): Path to a file with scheduler information if available.
        direct_to_workers (Optional[bool]): Whether or not to connect directly to the workers, or
            to ask the scheduler to serve as intermediary.
        heartbeat_interval (Optional[int]): Time in milliseconds between heartbeats to scheduler.
    '''

    def __new__(
        cls,
        address=None,
        timeout=None,
        scheduler_file=None,
        direct_to_workers=False,
        heartbeat_interval=None,
    ):
        return super(DaskConfig, cls).__new__(
            cls,
            address=check.opt_str_param(address, 'address'),
            timeout=check.opt_int_param(timeout, 'timeout'),
            scheduler_file=check.opt_str_param(scheduler_file, 'scheduler_file'),
            direct_to_workers=check.opt_bool_param(direct_to_workers, 'direct_to_workers'),
            heartbeat_interval=check.opt_int_param(heartbeat_interval, 'heartbeat_interval'),
        )

    @staticmethod
    def get_engine():
        from .engine import DaskEngine

        return DaskEngine

    def build_dict(self, pipeline_name):
        '''Returns a dict we can use for kwargs passed to dask client instantiation.

        Intended to be used like:

        with dask.distributed.Client(**cfg.build_dict()) as client:
            << use client here >>

        '''
        dask_cfg = {
            'name': pipeline_name,
        }

        # if address is set, don't add LocalCluster args
        # context: https://github.com/dask/distributed/issues/3313
        if not getattr(self, 'address', None):
            # We set threads_per_worker because Dagster is not thread-safe. Even though
            # processes=True by default, there is a clever piece of machinery
            # (dask.distributed.deploy.local.nprocesses_nthreads) that automagically makes execution
            # multithreaded by default when the number of available cores is greater than 4.
            # See: https://github.com/dagster-io/dagster/issues/2181
            # We may want to try to figure out a way to enforce this on remote Dask clusters against
            # which users run Dagster workloads.
            dask_cfg['threads_per_worker'] = 1

        for cfg_param in [
            'address',
            'timeout',
            'scheduler_file',
            'direct_to_workers',
            'heartbeat_interval',
        ]:
            cfg_value = getattr(self, cfg_param, None)
            if cfg_value:
                dask_cfg[cfg_param] = cfg_value
        return dask_cfg



import argparse
import sys

from setuptools import find_packages, setup


def get_version(name):
    version = {}
    with open('dagster_dask/version.py') as fp:
        exec(fp.read(), version)  # pylint: disable=W0122

    if name == 'dagster-dask':
        return version['__version__']
    elif name == 'dagster-dask-nightly':
        return version['__nightly__']
    else:
        raise Exception('Shouldn\'t be here: bad package name {name}'.format(name=name))


parser = argparse.ArgumentParser()
parser.add_argument('--nightly', action='store_true')


def _do_setup(name='dagster-dask'):
    setup(
        name=name,
        version=get_version(name),
        author='Elementl',
        license='Apache-2.0',
        description='Package for using Dask as Dagster\'s execution engine.',
        url='https://github.com/dagster-io/dagster/tree/master/python_modules/libraries/dagster-dask',
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(exclude=['test']),
        install_requires=[
            'bokeh',
            'dagster',
            'dagster_graphql',
            'dask>=1.2.2',
            'distributed>=1.28.1',
        ],
        zip_safe=False,
    )


if __name__ == '__main__':
    parsed, unparsed = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + unparsed

    if sys.version_info[0] < 3 and sys.platform == 'darwin':
        raise Exception('dagster-dask does not work on Python 2 + macOS')

    if parsed.nightly:
        _do_setup('dagster-dask-nightly')
    else:
        _do_setup('dagster-dask')



from dagster_graphql.client.query import START_PIPELINE_EXECUTION_MUTATION, SUBSCRIPTION_QUERY
from dagster_graphql.implementation.context import DagsterGraphQLContext
from dagster_graphql.implementation.pipeline_execution_manager import SubprocessExecutionManager
from dagster_graphql.schema import create_schema
from dagster_graphql.test.utils import execute_dagster_graphql
from graphql import graphql
from graphql.execution.executors.sync import SyncExecutor

from dagster import ExecutionTargetHandle
from dagster.core.instance import DagsterInstance
from dagster.utils import file_relative_path


def test_execute_hammer_through_dagit():
    handle = ExecutionTargetHandle.for_pipeline_python_file(
        file_relative_path(__file__, '../../../../examples/dagster_examples/toys/hammer.py'),
        'hammer_pipeline',
    )
    instance = DagsterInstance.local_temp()

    execution_manager = SubprocessExecutionManager(instance)

    context = DagsterGraphQLContext(
        handle=handle, execution_manager=execution_manager, instance=instance
    )

    executor = SyncExecutor()

    variables = {
        'executionParams': {
            'environmentConfigData': {'storage': {'filesystem': {}}, 'execution': {'dask': {}}},
            'selector': {'name': handle.build_pipeline_definition().name},
            'mode': 'default',
        }
    }

    start_pipeline_result = graphql(
        request_string=START_PIPELINE_EXECUTION_MUTATION,
        schema=create_schema(),
        context=context,
        variables=variables,
        executor=executor,
    )

    run_id = start_pipeline_result.data['startPipelineExecution']['run']['runId']

    context.execution_manager.join()

    subscription = execute_dagster_graphql(context, SUBSCRIPTION_QUERY, variables={'runId': run_id})

    subscribe_results = []
    subscription.subscribe(subscribe_results.append)

    messages = [x['__typename'] for x in subscribe_results[0].data['pipelineRunLogs']['messages']]

    assert 'PipelineStartEvent' in messages
    assert 'PipelineSuccessEvent' in messages



from dagster_dask.engine import get_dask_resource_requirements

from dagster import solid


def test_resource_tags():
    @solid(tags={'dagster-dask/resource_requirements': {'GPU': 1, 'MEMORY': 10e9}})
    def boop(_):
        pass

    reqs = get_dask_resource_requirements(boop.tags)
    assert reqs['GPU'] == 1
    assert reqs['MEMORY'] == 10e9
